Public Delegate Sub ConnectedEvent(ByVal Sender As Object, ByVal [Error] As Integer)
Public Delegate Sub DisconnectedEvent(ByVal Sender As Object, ByVal Reason As Integer)
Public Delegate Sub ValueChangedEvent(ByVal Sender As Object, ByVal Handle As UInt16, ByVal Value As Byte())
Public Delegate Sub WriteCompleted(ByVal Sender As Object, ByVal Result As Integer)

Public Class GattThread
    Inherits wclThread

    Private Const CMD_SEND_DATA = 1

    Private FAddress As Int64
    Private FRadio As wclBluetoothRadio

    Private WithEvents FClient As wclGattClient
    Private WithEvents FReceiver As wclMessageReceiver

    Private FData As Byte()
    Private FNotifyChar As wclGattCharacteristic
    Private FWriteChar As wclGattCharacteristic

    Private Sub InternalDisconnect(ByVal Reason As Integer)
        MyBase.Terminate()
        FReceiver.Close()

        If FData IsNot Nothing Then
            DoWriteCompleted(wclConnectionErrors.WCL_E_CONNECTION_TERMINATED_BY_USER)
            FData = Nothing
        End If
        DoDisconnected(wclConnectionErrors.WCL_E_CONNECTION_TERMINATED_BY_USER)

        FAddress = 0
        FRadio = Nothing
        FReceiver = Nothing
    End Sub

    Private Sub FReceiver_OnMessage(ByVal Message As wclMessage) Handles FReceiver.OnMessage
        If Message.Category = GattThreadMessage.MSG_CAT_GATT_THREAD Then
            Select Case Message.Id
                Case GattThreadMessage.MSG_ID_GATT_THREAD_CONNECTED
                    Dim ConnectMsg As GattThreadConnectedMessage = CType(Message, GattThreadConnectedMessage)
                    If ConnectMsg.[Error] <> wclErrors.WCL_E_SUCCESS Then
                        Terminate()
                    End If
                    DoConnected(ConnectMsg.[Error])

                Case GattThreadMessage.MSG_ID_GATT_THREAD_DISCONNECTED
                    Dim DiscMsg As GattThreadDisconnectedMessage = CType(Message, GattThreadDisconnectedMessage)
                    InternalDisconnect(DiscMsg.Reason)

                Case GattThreadMessage.MSG_ID_GATT_THREAD_VALUE_CHANGED
                    Dim DataMsg As GattThreadValueChangedMessage = CType(Message, GattThreadValueChangedMessage)
                    DoValueChanged(DataMsg.Handle, DataMsg.Value)

                Case GattThreadMessage.MSG_ID_GATT_THREAD_WRITE_COMPLETED
                    Dim WriteMsg As GattThreadWriteCompletedMessage = CType(Message, GattThreadWriteCompletedMessage)
                    FData = Nothing
                    DoWriteCompleted(WriteMsg.Result)
            End Select
        End If
    End Sub

    Private Sub FClient_OnCharacteristicChanged(ByVal Sender As Object, ByVal Handle As UShort, ByVal Value() As Byte) Handles FClient.OnCharacteristicChanged
        Dim Msg As GattThreadValueChangedMessage = New GattThreadValueChangedMessage(Handle, Value)
        Try
            FReceiver.Post(Msg)
        Finally
            Msg.Release()
        End Try
    End Sub

    Private Sub FClient_OnConnect(ByVal Sender As Object, ByVal [Error] As Integer) Handles FClient.OnConnect
        If [Error] = wclErrors.WCL_E_SUCCESS Then
            Dim Services As wclGattService() = Nothing
            Dim Res As Integer = FClient.ReadServices(wclGattOperationFlag.goNone, Services)
            If Res = wclErrors.WCL_E_SUCCESS And Services IsNot Nothing Then
                For Each Service As wclGattService In Services
                    Dim Chars As wclGattCharacteristic() = Nothing
                    Res = FClient.ReadCharacteristics(Service, wclGattOperationFlag.goNone, Chars)
                    If Res = wclErrors.WCL_E_SUCCESS And Chars IsNot Nothing Then
                        For Each [Char] As wclGattCharacteristic In Chars
                            If [Char].IsWritable And FWriteChar.Handle = 0 Then
                                FWriteChar = [Char]
                            End If

                            If ([Char].IsIndicatable Or [Char].IsNotifiable) And FNotifyChar.Handle = 0 Then
                                FNotifyChar = [Char]
                            End If

                            If FNotifyChar.Handle <> 0 And FWriteChar.Handle <> 0 Then
                                Exit For
                            End If
                        Next
                    End If

                    If FNotifyChar.Handle <> 0 And FWriteChar.Handle <> 0 Then
                        Exit For
                    End If
                Next
            End If
        End If

        If FNotifyChar.Handle <> 0 Then
            FClient.Subscribe(FNotifyChar)
            FClient.WriteClientConfiguration(FNotifyChar, True, wclGattOperationFlag.goNone)
        End If

        Dim Msg As GattThreadConnectedMessage = New GattThreadConnectedMessage([Error])
        Try
            FReceiver.Post(Msg)
        Finally
            Msg.Release()
        End Try
    End Sub

    Private Sub FClient_OnDisconnect(ByVal Sender As Object, ByVal Reason As Integer) Handles FClient.OnDisconnect
        Dim Msg As GattThreadDisconnectedMessage = New GattThreadDisconnectedMessage(Reason)
        Try
            FReceiver.Post(Msg)
        Finally
            Msg.Release()
        End Try
    End Sub

    Protected Overrides Function OnInitialize() As Boolean
        FClient = New wclGattClient()
        FClient.Address = FAddress
        Dim Res As Boolean = (FClient.Connect(FRadio) = wclErrors.WCL_E_SUCCESS)
        If Not Res Then
            FClient = Nothing
        End If
        Return Res
    End Function

    Protected Overrides Sub OnSignal(ByVal Id As UInteger)
        If Id = CMD_SEND_DATA And FData IsNot Nothing And FWriteChar.Handle <> 0 Then
            Dim Res As Integer = FClient.WriteCharacteristicValue(FWriteChar, FData, wclGattProtectionLevel.plNone)
            Dim Msg As GattThreadWriteCompletedMessage = New GattThreadWriteCompletedMessage(Res)
            Try
                FReceiver.Post(Msg)
            Finally
                Msg.Release()
            End Try
        End If
    End Sub

    Protected Overrides Sub OnTerminate()
        If FNotifyChar.Handle <> 0 Then
            FClient.WriteClientConfiguration(FNotifyChar, False, wclGattOperationFlag.goNone)
            FClient.Unsubscribe(FNotifyChar)
        End If

        FNotifyChar.Handle = 0
        FWriteChar.Handle = 0

        FClient.Disconnect()
        FClient = Nothing
    End Sub

    Protected Sub DoConnected(ByVal [Error] As Integer)
        RaiseEvent OnConnected(Me, [Error])
    End Sub

    Protected Sub DoDisconnected(ByVal Reason As Integer)
        RaiseEvent OnDisconnected(Me, Reason)
    End Sub

    Protected Sub DoValueChanged(ByVal Handle As UInt16, ByVal Value As Byte())
        RaiseEvent OnValueChanged(Me, Handle, Value)
    End Sub

    Protected Sub DoWriteCompleted(ByVal Result As Integer)
        RaiseEvent OnWriteCompleted(Me, Result)
    End Sub

    Public Sub New()
        FAddress = 0
        FRadio = Nothing

        FClient = Nothing
        FReceiver = Nothing

        FData = Nothing
        FNotifyChar.Handle = 0
        FWriteChar.Handle = 0
    End Sub

    Public Function Connect(ByVal Address As Int64, ByVal Radio As wclBluetoothRadio) As Integer
        If Radio Is Nothing Or Address = 0 Then
            Return wclErrors.WCL_E_SUCCESS
        End If

        If FRadio IsNot Nothing Then
            Return wclConnectionErrors.WCL_E_CONNECTION_ACTIVE
        End If

        FAddress = Address
        FRadio = Radio
        FReceiver = New wclMessageReceiver()

        Dim Res As Integer = FReceiver.Open()
        If Res = wclErrors.WCL_E_SUCCESS Then
            Res = MyBase.Run()
            If Res <> wclErrors.WCL_E_SUCCESS Then
                FReceiver.Close()
            End If
        End If

        If Res <> wclErrors.WCL_E_SUCCESS Then
            FAddress = 0
            FRadio = Nothing
            FReceiver = Nothing
        End If

        Return Res
    End Function

    Public Function Disconnect() As Integer
        If FRadio Is Nothing Then
            Return wclConnectionErrors.WCL_E_CONNECTION_CLOSED
        End If

        InternalDisconnect(wclConnectionErrors.WCL_E_CONNECTION_TERMINATED_BY_USER)

        Return wclErrors.WCL_E_SUCCESS
    End Function

    Public Function Write(ByVal Data As Byte())
        If Data Is Nothing Then
            Return wclErrors.WCL_E_INVALID_ARGUMENT
        End If

        If FData IsNot Nothing Then
            Return wclBluetoothErrors.WCL_E_BLUETOOTH_PENDING
        End If

        If FWriteChar.Handle = 0 Then
            Return wclBluetoothErrors.WCL_E_BLUETOOTH_LE_ATTRIBUTE_NOT_FOUND
        End If

        FData = Data
        Dim Res As Integer = Signal(CMD_SEND_DATA)
        If Res <> wclErrors.WCL_E_SUCCESS Then
            FData = Nothing
        End If
        Return Res
    End Function

    Public ReadOnly Property Address() As Int64
        Get
            Return FAddress
        End Get
    End Property

    Public ReadOnly Property Radio() As wclBluetoothRadio
        Get
            Return FRadio
        End Get
    End Property

    Public Event OnConnected As ConnectedEvent
    Public Event OnDisconnected As DisconnectedEvent
    Public Event OnValueChanged As ValueChangedEvent
    Public Event OnWriteCompleted As WriteCompleted
End Class
